﻿using System;

using Adwk.Mileage.Dac;
using Adwk.Mileage.Dac.AdwkMileagesDataSetTableAdapters;
using Adwk.Mileage.Data;
using System.Data.SqlClient;

namespace Adwk.Mileage.Logic
{
    public static class BLComp1
    {
        // このコンポーネントが返すメッセージ
        const string msgInvalidMile = "加算マイルは1以上である必要があります。";
        const string msgInvalidLevel = "加算のレベル{0}は正しくありません。";
        const string msgMemberNotFound = "指定した会員コードは存在しません。";
        const string msgMemberNull = "会員コードに null を指定できません。";
        const string msgDBError = "システムエラーです。(0701)";
        const string msgEtcError = "システムエラーです。(0702)";

        // 乗船などのマイル加算
        public static bool AddMileByMember(
            string memberCode,
            DateTime boardingDate, int baseMile, string mileageLevel, string cmt,
            out int appliedMile,
            out AdwkMStatus stat, out string systemMessage)
        {
            const string SecondClass = "2";  //二等など通常のクラスの加算レベル
            const string FirstClass = "1";   //一等など上位のクラスの加算レベル
            const int SecondRatio = 1;       //通常のクラスの倍率
            const int FirstRatio = 2;        //上位のクラスの倍率
            const string addedCmt = "[加算]";

            appliedMile = 0;

            // 会員コードの検証
            if(memberCode == null)
            {
                stat = AdwkMStatus.InvalidArgument;
                systemMessage = msgMemberNull;
                return false;
            }

            // マイル値の検証
            if(baseMile <= 0)
            {
                stat = AdwkMStatus.MileInvalid;
                systemMessage = msgInvalidMile;
                return false;
            }

            // 加算のレベル(乗船のクラス)から倍率を算出し、加算マイルを決定する
            int mile;  //加算マイル
            int ratio; //倍率
            switch(mileageLevel)
            {
                case SecondClass: ratio = SecondRatio; break;
                case FirstClass:  ratio = FirstRatio;  break;
                default:
                    stat = AdwkMStatus.InvalidLevel;
                    systemMessage = String.Format(msgInvalidLevel, mileageLevel);
                    return false;
            }
            mile = baseMile * ratio;
            appliedMile = mile;

            // 摘要欄に加算の旨を追加する
            // cmtはnull可、""に置き換える
            cmt = addedCmt + (cmt ?? "");

            // マイル加算の記録
            bool result = DAHelper1.RecordMileByMember(
                            memberCode, boardingDate, mile, cmt, out stat);
            switch (stat)
            {
                case AdwkMStatus.Success:
                    systemMessage = "";
                    break;
                case AdwkMStatus.MemberNotFound:
                    systemMessage = msgMemberNotFound;
                    break;
                case AdwkMStatus.DBError:
                    systemMessage = msgDBError;
                    break;
                default:
                    systemMessage = msgEtcError;
                    break;
            }
            return result;
        }

        // 会員名、マイル累計などの基本情報の取得
        //   このコンポーネントでは、汎用性を持たせて、
        // ユーザーインターフェイスのデータバインドで使用する既存データセットに対して
        // データを提供できるようにするため、呼び出し側が作成したDataTableに対して、
        // データを追加するようにしています。
        public static bool QueryInfo1ByMember(
            string memberCode, 
            AdwkMileagesDataSet.会員マイル累計DataTable table,
            out AdwkMStatus stat, out string systemMessage)
        {
            // 会員コードの検証
            if (memberCode == null)
            {
                stat = AdwkMStatus.InvalidArgument;
                systemMessage = msgMemberNull;
                return false;
            }

            var adapter = new 会員マイル累計TableAdapter();
            try
            {
                int count = adapter.Fill(table, memberCode);
                if (count < 1)
                {
                    stat = AdwkMStatus.MemberNotFound;
                    systemMessage = msgMemberNotFound;
                    return false;
                }
            }
            catch (SqlException)
            {
                stat = AdwkMStatus.DBError;
                systemMessage = msgDBError;
                return false;
            }
            catch
            {
                stat = AdwkMStatus.UndefinedError;
                systemMessage = msgEtcError;
                return false;
            }

            stat = AdwkMStatus.Success;
            systemMessage = "";
            return true;
        }

        // 会員のマイル履歴の取得
        //   このコンポーネントでは、汎用性を持たせて、
        // ユーザーインターフェイスのデータバインドで使用する既存データセットに対して
        // データを提供できるようにするため、呼び出し側が作成したDataTableに対して、
        // データを追加するようにしています。
        public static bool QueryHistoryByMemberAndPeriod(
            string memberCode, DateTime fromDate, DateTime untilDate,
            AdwkMileagesDataSet.会員マイル履歴DataTable table,
            out AdwkMStatus stat, out string systemMessage)
        {
            // 会員コードの検証
            if (memberCode == null)
            {
                stat = AdwkMStatus.InvalidArgument;
                systemMessage = msgMemberNull;
                return false;
            }

            var adapter = new 会員マイル履歴TableAdapter();

            try
            {
                adapter.Fill(table, memberCode, fromDate, untilDate);
                // 照会した結果がゼロ件の場合は、
                // 指定期間に履歴がないだけでなく、
                // 単に memberCode が間違っていた場合にゼロ件の場合もあります。
                // ここでは、memberCodeの有効性は確認しません。
                // いずれのゼロ件の場合も、該当無しとして正常終了します。
            }
            catch (SqlException)
            {
                stat = AdwkMStatus.DBError;
                systemMessage = msgDBError;
                return false;
            }
            catch
            {
                stat = AdwkMStatus.UndefinedError;
                systemMessage = msgEtcError;
                return false;
            }

            stat = AdwkMStatus.Success;
            systemMessage = "";
            return true;
        }
    }
}
