﻿using System;
using System.Transactions;

using Adwk.Mileage.Dac;
using Adwk.Mileage.Dac.AdwkMileagesDataSetTableAdapters;
using Adwk.Mileage.Data;

using Adwk.Mileage.Logic.ServiceReference1;

namespace Adwk.Mileage.Logic
{
    public static class BLComp2
    {
        // このコンポーネントが返すメッセージ
        const string msgInvalidMile = "移行マイルは1以上である必要があります。";
        const string msgMileNotEnough = "マイル残高が不足しています。";
        const string msgMemberNotFound = "指定した会員コードは存在しません。";
        const string msgMemberNull = "会員コード、または移行先の顧客IDに null を指定できません。";
        const string msgDBError = "システムエラーです。(0801)";
        const string msgEtcError = "システムエラーです。(0802)";
        const string msgExternalError = "移行先システムでエラーが発生しました。メッセージ: ";
        const string msgTrannsferSuccess = "移行手続きが完了しました。移行処理番号:{0}";

        // (特定の)外部ポイント管理システムへのマイル移行
        public static bool TransferMileByMember(
            string memberCode,
            string toCustomerId, DateTime appliedDate, int mile, 
            out AdwkMStatus stat, out string systemMessage)
        {
            const string fromSystem   = "AWTS";  //外部サービスに伝える転送元の名称
            const string transferdCmt = "[移行] 移行先の顧客ID:{0}";  //社内記録データの摘要欄

            // 会員コード、顧客IDの検証
            if (memberCode == null || toCustomerId == null)
            {
                stat = AdwkMStatus.InvalidArgument;
                systemMessage = msgMemberNull;
                return false;
            }

            // マイル値の検証(1以上であること)
            if (mile <= 0)
            {
                stat = AdwkMStatus.MileInvalid;
                systemMessage = msgInvalidMile;
                return false;
            }

            // 摘要欄に移行の旨を追加する
            // cmtはnull可、""に置き換える
            string cmt = String.Format(transferdCmt, toCustomerId);

            // 社内マイル減算、および、外部へのマイル移行の分散トランザクション
            using (TransactionScope ts =
                   new TransactionScope(TransactionScopeOption.Required))
            {
                bool result;

                // 社内マイル減算を記録する
                int subMile = -mile;
                result = DAHelper1.RecordMileByMember(
                             memberCode, appliedDate, subMile, cmt, out stat);
                if (!result)
                {
                    switch (stat)
                    {
                        case AdwkMStatus.Success:
                            systemMessage = "";
                            break;
                        case AdwkMStatus.MemberNotFound:
                            systemMessage = msgMemberNotFound;
                            break;
                        case AdwkMStatus.DBError:
                            systemMessage = msgDBError;
                            break;
                        case AdwkMStatus.MileNotEnough:
                            systemMessage = msgMileNotEnough;
                            break;
                        default:
                            systemMessage = msgEtcError;
                            break;
                    }
                    // 記録失敗のため処理中止(SetCompleteしない)
                    return false;
                }

                // この時点で、社内の記録は暫定的に正常終了
                System.Diagnostics.Debug.Assert(
                    result && (stat == AdwkMStatus.Success));
                systemMessage = "";
                
                // 社外移行のための名前取得(既存のDACを流用)
                // (この移行では、移行先と名義が一致しないと移行できない)
                string toLastName, toFirstName;
                var table1 = new AdwkMileagesDataSet.会員マイル累計DataTable();
                var adapter1 = new 会員マイル累計TableAdapter();
                int count = adapter1.Fill(table1, memberCode);
                System.Diagnostics.Debug.Assert(count > 0); //すでに該当会員がいることはチェック済み
                toLastName = table1[0].姓;
                toFirstName = table1[0].名;

                // 社外のポイント管理システムへの移行
                int transferId;
                string externalMessage;
                var proxy = new ExternalServiceClient();
                result = proxy.TransferPoint(
                            out transferId, out externalMessage,
                            fromSystem, 
                            toCustomerId, toLastName, toFirstName, mile);
                if (!result)
                {
                    systemMessage = msgExternalError + (externalMessage ?? "");
                    stat = AdwkMStatus.ExternalError;
                    // 外部システムでのエラーのため処理中止(SetCompleteしない)
                    return false;
                }

                // この時点で、両方の処理が成功
                systemMessage = String.Format(msgTrannsferSuccess, transferId);
                stat = AdwkMStatus.Success;

                //◆◆実験用コード◆◆
                //  移行マイル(移行ポイント)が「77」マイルの場合は、SetCompleteせずに終了。
                //  外部サービスもロールバックできる。
                if (mile == 77)
                {
                    systemMessage = "ロールバックの実験";
                    stat = AdwkMStatus.UndefinedError;
                    return false;
                }

                //正常終了に投票(vote)
                ts.Complete();
            }
            
            // 正常終了
            return true;
        }
    }
}
