﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Transactions;

using Adwk.Reservation.Dac;
using Adwk.Reservation.Data;
using Adwk.Reservation.Logic.ServiceReference1;

namespace Adwk.Reservation.Logic
{
    // 乗船手続きのコンポーネント
    //   ここでは簡単にするため、予約コードから予約データを照会し確認したのち、
    // そのデータについて乗船手続きを行うのではなく、乗船手続きで必要な情報を
    // 一通り引数で指定して手続き処理を行うようにしています。
    // このため、ステートレスなコンポーネントでよいので、静的クラスを使用しています。
    public static class BLComp4
    {
        // このコンポーネントが返すメッセージ
        const string msgInvalidReservationCode = "指定した予約コードが正しくありません。";
        const string msgNoReservation = "指定した予約は存在しません。";
        const string msgAlreadyCheckedIn = "既に乗船手続きが済んでいます。";
        const string msgCheckedIn = "乗船手続きに成功しました。";
        const string msgCheckedInButMileageErr = "[注意] 乗船手続きに成功しましたが、マイル加算に失敗しました。" +
                                                 "マイル管理サービスからのメッセージ: ";
        const string msgCheckedInAndMileageAdded = "乗船手続きとマイル加算に成功しました。";
        const string msgReservationDbAccessError = "システムエラーです。(0501)";
        const string msgReservationEtcError = "システムエラーです。(0502)";


        // 乗船の手続き
        //   指定された運行コード(scheduleCode)とクラス(levelCode)に対して
        //   指定された予約コード(reservationCode)の乗客(lastName、firstName)に関する
        //   チェックイン(乗船手続き)を行います。
        //   (乗船にはあらかじめ予約が必要です。)
        //   指定された情報に一致する予約がない場合、手続きは失敗します。
        //   この時点で会員コードを指定することで、既存の予約データの会員コードを上書きすることができます。
        public static bool CheckIn(
            string scheduleCode, string levelCode,
            string reservationCode, string memberCode,
            string lastName, string firstName,
            out AdwkRStatus stat, out string systemMessage)
        {
            const string checkedInCmt = "(乗船済み)";

            // 予約コードの検証(数値化できるか確認)
            int reservationIntCode = 0;
            try 
            { 
                reservationIntCode = Convert.ToInt32(reservationCode); 
            }
            catch
            {
                stat = AdwkRStatus.InvalidArgument;
                systemMessage = msgInvalidReservationCode;
                return false;
            }

            // 乗船手続きのほか、マイルの登録も行う
            // (乗船手続きとマイル登録は別のトランザクション)
            var oc = new AdwkReservationsEntities();
            予約 rsv1 = null;
            try
            {
                // 乗船手続き
                using (var ts = new TransactionScope(TransactionScopeOption.RequiresNew))
                {
                    // 予約データを取得
                    var query1 = from r in oc.予約
                                 where r.予約コード == reservationIntCode &&
                                       r.運行コード == scheduleCode &&
                                       r.クラスコード == levelCode &&
                                       r.姓 == lastName && r.名 == firstName
                                 select r;
                    try
                    {
                        rsv1 = query1.First();
                    }
                    catch (InvalidOperationException)
                    {
                        stat = AdwkRStatus.NoReservation;
                        systemMessage = msgNoReservation;
                        return false;
                    }
                    catch
                    {
                        stat = AdwkRStatus.UndefinedError;
                        systemMessage = msgReservationDbAccessError;
                        return false;
                    }
                    System.Diagnostics.Debug.Assert(rsv1 != null);
                    
                    // 乗船手続きが行われていない状態であるかチェック
                    if (rsv1.乗船受付フラグ != AdwkCheckInFlag.NotCheckedIn)
                    {
                        stat = AdwkRStatus.AleadyCheckedIn;
                        systemMessage = msgAlreadyCheckedIn;
                        return false;
                    }

                    // 会員コードを改めて指定した場合は、会員コードを置き換え
                    // ここでは簡単にするため、会員コードの有効性検証を行いません。)
                    if (!String.IsNullOrEmpty(memberCode))
                    {
                        rsv1.会員コード = memberCode;
                    }

                    // 予約データ更新
                    rsv1.コメント += checkedInCmt;
                    rsv1.乗船受付フラグ = AdwkCheckInFlag.CheckedIn;
                    rsv1.乗船日時 = rsv1.運行予定明細.運行予定.運行日;
                    //rsv1.乗船日時 = DateTime.Now; 

                    // Note.
                    // 乗船日時は、サンプル操作を行なうシステム時刻ではなく、
                    // 実際に乗船予定の日になるようにするため、DateTime.Now を
                    // 使わずに、運航日をそのまま乗船日とします。
                    // この日付けはマイル加算日になっています。

                    // 乗船手続きによる予約データの更新
                    oc.SaveChanges();

                    // 正常終了に投票(Vote)
                    ts.Complete();

                    // 正常終了の情報設定
                    stat = AdwkRStatus.Success;
                    systemMessage = msgCheckedIn;
                }
                System.Diagnostics.Debug.Assert(rsv1 != null);

                // 会員コードが null の場合、これ以上は処理をしない
                if (rsv1.会員コード == null)
                {
                    stat = AdwkRStatus.Success;
                    systemMessage = msgCheckedIn;
                    return true;
                }

                // マイル加算処理
                //   上記のトランザクションとまとめられたときに、1つのトランザクションに
                // ならないようにするため、Requires.New を使用する
                using (TransactionScope ts = new TransactionScope(TransactionScopeOption.RequiresNew))
                {
                    // マイル加算に必要な基本マイルを求める
                    int baseMile;
                    var query2 = from rt in oc.路線情報
                                 where rt.出発地コード == rsv1.運行予定明細.運行予定.便情報.出発地コード &&
                                       rt.到着地コード == rsv1.運行予定明細.運行予定.便情報.到着地コード
                                 select rt.運行マイル;
                    baseMile = query2.First();  //有効な予約に基づく検索なので、該当データはあるはず

                    // マイル管理サービスの呼び出しによる加算
                    var proxy = new MileageUpdateClient();
                    bool result;
                    int appliedMile;
                    AdwkMStatus mstat;
                    string mmsg;
                    result = proxy.AddMileByMember(
                                out appliedMile, out mstat, out mmsg,
                                rsv1.会員コード, (DateTime)rsv1.乗船日時,
                                baseMile, rsv1.クラスコード, rsv1.コメント);
                    if (!result)
                    {
                        // 乗船手続きに成功したが、マイル加算に失敗した場合は、
                        // 注意を促すメッセージを作成するが、乗船手続き自体は
                        // 成功したので、true を返す
                        stat = AdwkRStatus.SuccessButMileage;
                        systemMessage =  msgCheckedInButMileageErr + mmsg;
                        return true;
                    }

                    // 予約データの更新
                    rsv1.実績マイル = appliedMile;

                    // データベース更新
                    oc.SaveChanges();
                
                    // 正常終了に投票(Vote)
                    ts.Complete();

                    // 正常終了の情報設定
                    stat = AdwkRStatus.Success;
                    systemMessage = msgCheckedInAndMileageAdded;
                }
                return true;
            }
            catch
            {
                stat = AdwkRStatus.UndefinedError;
                systemMessage = msgReservationEtcError;
                return false;
            }
            finally
            {
                if (oc != null) oc.Dispose();
            }
        }
    }
}
