﻿using System;
using System.ServiceModel;

using Adwk.Reservation.Data;
using Adwk.Reservation.Logic;

namespace Adwk.Reservation.Service
{
    // 予約管理サービス
    // (予定照会、予約、予約照会、予約削除)
    //   インスタンス単位(セッション単位)で、確認 → 削除 という流れが
    // あるので、インスタンス単位のシングルスレッドに設定します。
    //   インスタンス内はスレッドセーフではないですが、インスタンスが異なれば、
    // メンバー変数も異なるので、インスタンスごとに並行実行できます。
    [ServiceBehavior(
        InstanceContextMode=InstanceContextMode.PerSession,
        ConcurrencyMode=ConcurrencyMode.Single)]
    public class ReservationService
        : IReservation
    {
        // このサービス固有のメッセージ
        const string msgGeneralError
            = "(内部エラー) システムメッセージ : {0}";

        // セッション単位のステートのためコンポーネントを維持
        private BLComp3 comp3 = null;  // 確認と削除

        // 運行予定の取得
        bool IReservation.QuerySchedule(
            DateTime fromDate, DateTime untilDate, 
            string deptCode, string arrvCode, string levelCode,
            out ScheduleInfo[] sinfos, out string systemMessage)
        {
            // 予定検索、予約用コンポーネントによる予定検索
            bool result;
            AdwkRStatus stat;
            string msg;
            result = BLComp2.QuerySchedule(
                fromDate, untilDate, deptCode, arrvCode, levelCode,
                out sinfos, out stat, out msg);

            // statは返さない
            // 返すメッセージの補正
            if (result) systemMessage = msg;
            else systemMessage = String.Format(msgGeneralError, msg);

            return result;
        }
        
        // 予約処理
        bool IReservation.Reserve(string scheduleCode, string levelCode, string memberCode, string lastName, string firstName, decimal appliedPrice, decimal appliedTax, string cmt, out string reservationCode, out string systemMessage)
        {
            // 予定検索、予約用コンポーネントによる予約
            bool result;
            AdwkRStatus stat;
            string msg;
            result = BLComp2.Reserve(
                scheduleCode, levelCode,
                memberCode, lastName, firstName,
                appliedPrice, appliedTax, cmt,
                out reservationCode, out stat, out msg);

            // statは返さない
            // 返すメッセージの補正
            if (result) systemMessage = msg;
            else systemMessage = String.Format(msgGeneralError, msg);

            return result;
        }

        // 予約の照会
        bool IReservation.QueryReservation1ByRCode(
            string reservationCode, string lastName, string firstName,
            out ReservationInfo reservation, out string systemMessage)
        {
            // ★テスト用★ Principalの確認
            //var curId = System.Threading.Thread.CurrentPrincipal.Identity;
            //Console.WriteLine("User:" + curId.Name );
            //Console.WriteLine("Type:" + curId.AuthenticationType);

            // コンポーネントがなければ作成
            if (comp3 == null) comp3 = new BLComp3();

            // 予約照会用コンポーネントによる照会
            bool result;
            AdwkRStatus stat;
            string msg;
            result = comp3.QueryReservation1ByRCode(
                reservationCode, lastName, firstName,
                out reservation, out stat, out msg);

            // statは返さない
            // 返すメッセージの補正
            if (result) systemMessage = msg;
            else systemMessage = String.Format(msgGeneralError, msg);

            return result;
        }

        // 予約の削除
        bool IReservation.DeleteReservationByRCode(
            string reservationCode, out string systemMessage)
        {
            // コンポーネントがなければ作成
            if (comp3 == null) comp3 = new BLComp3();

            // 予約照会用コンポーネントによる予約削除
            bool result;
            AdwkRStatus stat;
            string msg;
            result = comp3.DeleteReservationByRCode(
                reservationCode, out stat, out msg);

            // statは返さない
            // 返すメッセージの補正
            if (result) systemMessage = msg;
            else systemMessage = String.Format(msgGeneralError, msg);

            return result;
        }

        // ステートの強制リセット
        void IReservation.Reset()
        {
            if (comp3 != null)
            {
                comp3.Dispose();
                comp3 = null;
            }
        }

        // (テスト用メソッド)サービスインスタンスのハッシュ値を調べる
        int IReservation.GetInstanceHashCode()
        {
            return this.GetHashCode();
        }
    }
}
