﻿using System;
using System.Linq;
using System.Transactions;

using Adwk.Reservation.Dac;
using Adwk.Reservation.Data;

namespace Adwk.Reservation.Logic
{
    // 予約関連のロジックを含むビジネスコンポーネント
    // (運行予定の照会と予約)
    // 　このコンポーネントでは、問い合わせの都度、最新情報を返し、
    // 指定された運行コードとクラスコードを元に予約登録をするので、
    // ステートを保持する必要はありません。よって、静的クラスにしています。
    public static class BLComp2
    {
        // このコンポーネントが返すメッセージ
        const string msgInvalidDeptCode = "出発地コードが正しくありません。";
        const string msgInvalidArrvCode = "到着地コードが正しくありません。";
        const string msgInvalidLevel = "クラスコード(一等、二等など)が正しくありません。";
        const string msgInvalidScheduleId = "運行コードが正しくありません。";
        const string msgNameMissing = "姓および名は必須項目です。";
        const string msgNoRoute = "指定された区間は運行しておりません。";
        const string msgNoSchedule = "指定区間は有効な運行区間ですが、指定条件での運用予定はありません。";
        const string msgNoRoom = "申し訳けございません。指定された便の空きが無くなりました。";
        const string msgSuccessQuerySchedule = "{0}件のデータが見つかりました。";
        const string msgSuccessReservation = "予約が正常に完了しました。";
        const string msgRouteDBAccessError = "システムエラーです。(0101)";
        const string msgScheduleDBAccessError = "システムエラーです。(0102)";
        const string msgReserveEtcError = "システムエラーです。(0103)";
        const string msgFareCalcEtcError = "システムエラーです。(0104)";

        // 運行予定の取得
        //   簡単にするため、引数 deptCode, arrvCode は必須
        //   引数level は "" や null を指定可能(ワイルドカード)
        public static bool QuerySchedule(
            DateTime fromDate, DateTime untilDate,
            string deptCode, string arrvCode, string level,
            out ScheduleInfo[] sinfos, // 配列に変更 out IEnumerable<ScheduleInfo> sinfos,
            out AdwkRStatus stat, out string systemMessage)
        {
            const string FirstClassCode = "1";     //一等のコード
            const decimal FirstExtraPrice = 5000m; //一等の割り増し料金
            const decimal TaxRatio = 0.05m;        //税率
            
            sinfos = null;
            systemMessage = null;

            // 引数の検証
            if (String.IsNullOrEmpty(deptCode)) //出発地コード
            {
                sinfos = new ScheduleInfo[0];
                stat = AdwkRStatus.InvalidArgument;
                systemMessage = msgInvalidDeptCode;
                return false;
            }
            if(String.IsNullOrEmpty(arrvCode))  //到着地コード
            {
                sinfos = new ScheduleInfo[0];
                stat = AdwkRStatus.InvalidArgument;
                systemMessage = msgInvalidDeptCode;
                return false;
            }
            // ここでは DAC やデータのエンティティとして、ObjectContextを使用
            var oc = new AdwkReservationsEntities();

            // 路線(区間)の有効性チェックと路線情報取得
            // (このロジックを実装できる前提として、引数deptCodeと引数arrvCodeを必須にして、
            //  路線が1つに特定できる必要があります。
            //  路線の引数にワイルドカードを許可する仕様にすると、路線が特定できません。)
            decimal basePrice;  //指定された路線の基本料金
            int     mile;       //指定された路線の運行マイル
            var query1 = from r in oc.路線情報
                         where r.出発地コード == deptCode && r.到着地コード == arrvCode
                         select new { r.基本料金, r.運行マイル };
            try
            {
                var routeInfo = query1.First();
                basePrice = routeInfo.基本料金;
                mile = routeInfo.運行マイル;
            }
            catch (InvalidOperationException) //区間が見つからない場合
            {
                stat = AdwkRStatus.NoRoute;
                systemMessage = msgNoRoute;
                return true;  //区間が見つからないのは例外ではなく正常終了扱い
            }
            catch
            {
                stat = AdwkRStatus.UndefinedError;
                systemMessage = msgRouteDBAccessError;
                return false;
            }
            // この時点で、basePrice(基本料金)、mile(運行マイル)が決定される             

            // 運行予定を取得するクエリ式の構築
            //   クエリ式を用いれば、WCFでのやり取りやコンポーネント間のやり取りで
            // 必要となるデータ型に簡単に変換できます。
            var query2 = from s in oc.運行予定
                         join c in oc.便情報 on s.便名 equals c.便名
                         join p1 in oc.場所 on c.出発地コード equals p1.場所コード
                         join p2 in oc.場所 on c.到着地コード equals p2.場所コード
                         join d in oc.運行予定明細 on s.運行コード equals d.運行コード
                         join cl in oc.クラス on d.クラスコード equals cl.クラスコード
                         where (s.運行日 >= fromDate && s.運行日 < untilDate) &&
                               (deptCode == c.出発地コード) &&
                               (arrvCode == c.到着地コード) &&
                               (String.IsNullOrEmpty(level) || level == d.クラスコード)
                         select new ScheduleInfo()
                         {
                             ScheduleCode = s.運行コード,
                             ScheduledDate = s.運行日,
                             CruiseId = s.便名,
                             DeptCode = c.出発地コード,
                             DeptName = p1.場所名,
                             DeptTime = c.出発時刻,
                             ArrvCode = c.到着地コード,
                             ArrvName = p2.場所名,
                             ArrvTime = c.到着時刻,
                             Price = basePrice,  //特定路線の基本料金
                             Mile =  mile,　     //特定路線の運行マイル
                             LevelCode = d.クラスコード,
                             LevelName = cl.クラス名,
                             Capacity = d.定員数,
                             Room = d.空き数,
                             // 適用価格の算出ロジック
                             AppliedPrice = (d.クラスコード == FirstClassCode) ?
                                             basePrice + FirstExtraPrice : basePrice,
                             AppliedTax = ((d.クラスコード == FirstClassCode) ?
                                             basePrice + FirstExtraPrice : basePrice) * TaxRatio
                         };
             
            // 遅延実行に関する各種実験

            // 配列化することで確実にクエリを実行できる
            try
            {
                // 運行予定の配列を求める(この時点でクエリ実行)
                // 該当データがない場合は要素ゼロ個の配列が返る
                sinfos = query2.ToArray();
                if (sinfos.Length <= 0)
                {
                    // 有効な区間だが、該当予定なし
                    // 戻り値は正常終了扱い
                    stat = AdwkRStatus.NoSchedule;
                    systemMessage = msgNoSchedule;
                }
                else
                {
                    //n件検出
                    stat = AdwkRStatus.Success;
                    systemMessage = String.Format(msgSuccessQuerySchedule, sinfos.Length);  
                }
            }
            catch
            {
                stat = AdwkRStatus.UndefinedError;
                systemMessage = msgScheduleDBAccessError;
                return false;
            }
            finally
            {
                // 遅延実行の場合であれば、リソースは破棄できない
                // 接続を必要とする操作では、インスタンスが必要
                oc.Dispose();   // 遅延実行の場合は、これは使用できない。ToArrayにすれば可能。

                // oc.Dispose()をせず、CGを実行しても、
                // クエリ式が参照される限り、ObjectContextは残っている
                // クエリ式が遅延実行できる
                // GC.Collect(); 
            }

            //oc.Dispose()しない場合の、遅延実行の各種実験

            // あらかじめダミーで読み込み操作を行っても、前述のクエリを再び実行する際に、
            // 接続は必要になる
            // var test = query.First();

            // これを用いると、Count()の呼び出しと、foreachループと2回呼び出してしまう。
            // int count = sinfos.Count();

            // 遅延実行であると、呼び出し元でデータベースの接続エラーが発生する場合がある
            // WCFで返す場合は、WCFでのシリアル化のタイミング

            // 正常終了
            System.Diagnostics.Debug.Assert(sinfos != null);
            System.Diagnostics.Debug.Assert(stat == AdwkRStatus.NoRoute    ||
                                            stat == AdwkRStatus.NoSchedule ||
                                            stat == AdwkRStatus.Success);
            System.Diagnostics.Debug.Assert(systemMessage != null);
            return true;
        }

        // 予約処理
        //   運行コード scheduleCode とクラスコード levelCode を指定することで、予約対象を特定できます。
        //   簡単にするため、必要な利用者(乗客)の情報は氏名だけです。
        //   会員コード memberCode は、"" や null も可能。
        //   簡単にするため、この時点では会員コードの有効性は検証しません。
        //   (会員コードは、別のマイレージ管理サービスの管理下にあります。)
        //   引数 appliedPrice と appliedTax は、最終的に予約時に決定された適用料金であり、
        //   呼び出し側が自由に設定できますが、簡単にするため、その有効性は検証しません。
        //   引数 cmt は、予約時に呼び出し元が任意に追加するコメント(摘要)
        public static bool Reserve(
            string scheduleCode, string levelCode,
            string memberCode, string lastName, string firstName,
            decimal appliedPrice, decimal appliedTax, string cmt,
            out string reservationCode, out AdwkRStatus stat, out string systemMessage)
        {
            const string reservedCmt = "(予約済み)";  // このコンポーネントが追加するコメント(摘要)

            reservationCode = null;
            systemMessage = null;

            // 引数の検証
            if (String.IsNullOrEmpty(lastName) ||
                String.IsNullOrEmpty(firstName))  //姓と名の有無確認
            {
                stat = AdwkRStatus.InvalidArgument;
                systemMessage = msgNameMissing;
                return false;
            }

            try
            {
                using (var ts = new TransactionScope(TransactionScopeOption.Required))
                {
                    // ここでは DAC やデータのエンティティとして、ObjectContextを使用
                    using (var oc = new AdwkReservationsEntities())
                    {
                        // 改めて空きがあるか確認(運行コードとクラスコードの有効性も検証)
                        運行予定明細 sinfo;
                        var query1 = from s in oc.運行予定明細
                                     where s.運行コード == scheduleCode && s.クラスコード == levelCode
                                     select s;
                        try
                        {
                            sinfo = query1.First(); // 対象の運行予定明細の1件分データ
                        }
                        catch (InvalidOperationException) // 運行コードまたはクラスコードが不正で該当データがない
                        {
                            stat = AdwkRStatus.InvalidArgument;
                            systemMessage = msgInvalidScheduleId;
                            return false;
                        }
                        catch //その他のアクセスエラー
                        {
                            stat = AdwkRStatus.UndefinedError;
                            systemMessage = msgScheduleDBAccessError;
                            return false;
                        }
                        if (sinfo.空き数 <= 0) // 空き数不足
                        {
                            stat = AdwkRStatus.NoRoom;
                            systemMessage = msgNoRoom;
                            return true; // 正常終了扱い
                        }
                        System.Diagnostics.Debug.Assert(sinfo.空き数 > 0);
                        // 空き数減算
                        sinfo.空き数--;

                        // 会員コードがある場合の有効性検証
                        // (ここでは簡単にするため行いません。)

                        // 指定された適用料金の有効性確認
                        // (ここでは簡単にするため行いません。)

                        // 予約日の設定
                        // Note..
                        // (本来の予約日は、予約した日とすべきですが、
                        //  このサンプルを実行する日が、運航日よりも
                        //  後の可能性があるので、予約日を運航日の3日前に
                        //  なるように算出します。)
                        // DateTime rDate = DateTime.Now;
                        DateTime rDate = sinfo.運行予定.運行日.AddDays(-3);

                        // 新規予約の追加
                        var rinfo = 予約.Create予約(
                                        0, lastName, firstName,
                                        sinfo.運行コード, sinfo.クラスコード,
                                        appliedPrice, appliedTax,
                                        cmt + reservedCmt,
                                        rDate, AdwkCheckInFlag.NotCheckedIn);
                        // 会員コードがあれば設定
                        if (!String.IsNullOrEmpty(memberCode))
                        {
                            rinfo.会員コード = memberCode;
                        }
                        // ObjectContextに予約を追加
                        oc.予約.AddObject(rinfo);

                        // データを反映
                        oc.SaveChanges();

                        // 新規予約に割り振られた予約コードを取得
                        reservationCode = rinfo.予約コード.ToString();
                    }

                    // 正常終了に投票(Vote)
                    ts.Complete();

                    // 正常終了の情報設定
                    stat = AdwkRStatus.Success;
                    systemMessage = msgSuccessReservation;
                }
            }
            catch
            {
                stat = AdwkRStatus.UndefinedError;
                systemMessage = msgReserveEtcError;
                return false;
            }

            // 正常終了
            System.Diagnostics.Debug.Assert(reservationCode != null);
            System.Diagnostics.Debug.Assert(stat == AdwkRStatus.Success);
            System.Diagnostics.Debug.Assert(systemMessage != null);
            return true;
        }

        // 予約処理(引数最小のバージョン、Webサイト用に追加)
        //   運行コード scheduleCode とクラスコード levelCode を指定することで、予約対象を特定できます。
        //   簡単にするため、必要な利用者(乗客)の情報は氏名だけです。
        //   会員コード memberCode は、"" や null も可能。
        //   簡単にするため、この時点では会員コードの有効性は検証しません。
        //   (会員コードは、別のマイレージ管理サービスの管理下にあります。)
        public static bool Reserve(
            string scheduleCode, string levelCode,
            string memberCode, string lastName, string firstName,
            out string reservationCode, out AdwkRStatus stat, out string systemMessage)
        {
            const string reservedCmt = "(Webで予約済み)";  // このコンポーネントが追加するコメント(摘要)

            reservationCode = null;
            systemMessage = null;

            // 引数の検証
            if (String.IsNullOrEmpty(lastName) ||
                String.IsNullOrEmpty(firstName))  //姓と名の有無確認
            {
                stat = AdwkRStatus.InvalidArgument;
                systemMessage = msgNameMissing;
                return false;
            }

            try
            {
                using (var ts = new TransactionScope(TransactionScopeOption.Required))
                {
                    // ここでは DAC やデータのエンティティとして、ObjectContextを使用
                    using (var oc = new AdwkReservationsEntities())
                    {
                        // 改めて空きがあるか確認(運行コードとクラスコードの有効性も検証)
                        運行予定明細 sinfo;
                        var query1 = from s in oc.運行予定明細
                                     where s.運行コード == scheduleCode && s.クラスコード == levelCode
                                     select s;
                        try
                        {
                            sinfo = query1.First(); // 対象の運行予定明細の1件分データ
                        }
                        catch (InvalidOperationException) // 運行コードまたはクラスコードが不正で該当データがない
                        {
                            stat = AdwkRStatus.InvalidArgument;
                            systemMessage = msgInvalidScheduleId;
                            return false;
                        }
                        catch //その他のアクセスエラー
                        {
                            stat = AdwkRStatus.UndefinedError;
                            systemMessage = msgScheduleDBAccessError;
                            return false;
                        }
                        if (sinfo.空き数 <= 0) // 空き数不足
                        {
                            stat = AdwkRStatus.NoRoom;
                            systemMessage = msgNoRoom;
                            return true; // 正常終了扱い
                        }
                        System.Diagnostics.Debug.Assert(sinfo.空き数 > 0);
                        // 空き数減算
                        sinfo.空き数--;

                        // 会員コードがある場合の有効性検証
                        // (ここでは簡単にするため行いません。)

                        // 運賃再計算
                        // (このバージョンのReserveメソッドでは、
                        //  予約時に適用料金を引数で受け取りません。
                        //  簡単にするため、ここで運賃を再計算します。)
                        const string FirstClassCode = "1";     //一等のコード
                        const decimal FirstExtraPrice = 5000m; //一等の割り増し料金
                        const decimal TaxRatio = 0.05m;        //税率
                        decimal appliedPrice;
                        decimal appliedTax;
                        var query2 = from r in oc.路線情報
                                     where r.出発地コード == sinfo.運行予定.便情報.出発地コード &&
                                           r.到着地コード == sinfo.運行予定.便情報.到着地コード
                                     select new { r.基本料金, r.運行マイル };
                        try
                        {
                            var routeInfo = query2.First();
                            var basePrice = routeInfo.基本料金;
                            appliedPrice = ( levelCode == FirstClassCode) ?
                                             basePrice + FirstExtraPrice : basePrice;
                            appliedTax = ((levelCode == FirstClassCode) ?
                                           basePrice + FirstExtraPrice : basePrice) * TaxRatio;
                        }
                        catch
                        {
                            stat = AdwkRStatus.UndefinedError;
                            systemMessage = msgFareCalcEtcError;
                            return false;
                        }

                        // 予約日の設定
                        // Note..
                        // (本来の予約日は、予約した日とすべきですが、
                        //  このサンプルを実行する日が、運航日よりも
                        //  後の可能性があるので、予約日を運航日の3日前に
                        //  なるように算出します。)
                        // DateTime rDate = DateTime.Now;
                        DateTime rDate = sinfo.運行予定.運行日.AddDays(-3);
                        
                        // 新規予約の追加
                        var rinfo = 予約.Create予約(
                                        0, lastName, firstName,
                                        sinfo.運行コード, sinfo.クラスコード,
                                        appliedPrice, appliedTax,
                                        reservedCmt,
                                        rDate, AdwkCheckInFlag.NotCheckedIn);
                        // 会員コードがあれば設定
                        if (!String.IsNullOrEmpty(memberCode))
                        {
                            rinfo.会員コード = memberCode;
                        }
                        // ObjectContextに予約を追加
                        oc.予約.AddObject(rinfo);

                        // データを反映
                        oc.SaveChanges();

                        // 新規予約に割り振られた予約コードを取得
                        reservationCode = rinfo.予約コード.ToString();
                    }

                    // 正常終了に投票(Vote)
                    ts.Complete();

                    // 正常終了の情報設定
                    stat = AdwkRStatus.Success;
                    systemMessage = msgSuccessReservation;
                }
            }
            catch
            {
                stat = AdwkRStatus.UndefinedError;
                systemMessage = msgReserveEtcError;
                return false;
            }

            // 正常終了
            System.Diagnostics.Debug.Assert(reservationCode != null);
            System.Diagnostics.Debug.Assert(stat == AdwkRStatus.Success);
            System.Diagnostics.Debug.Assert(systemMessage != null);
            return true;
        }
    }
}

/* 出発地や到着地に "" または null を指定できるワイルドカードバージョン */
/*
            // 運行予定を取得するクエリ式の構築
            //   クエリ式を用いれば、WCFでのやり取りやコンポーネント間のやり取りで
            // 必要となるデータ型に簡単に変換できる
            var oc = new AdwkReservationsEntities();
            var query3 = from s in oc.運行予定
                         join c in oc.便情報 on s.便名 equals c.便名
                         join p1 in oc.場所 on c.出発地コード equals p1.場所コード
                         join p2 in oc.場所 on c.到着地コード equals p2.場所コード
                         join r in oc.路線情報 on c.出発地コード + c.到着地コード
                                             equals r.出発地コード + r.到着地コード  //単純にキーが連結するだけなので、可変の場合は注意
                         join d in oc.運行予定明細 on s.運行コード equals d.運行コード
                         join cl in oc.クラス on d.クラスコード equals cl.クラスコード
                         where (s.運行日 >= fromDate && s.運行日 < untilDate) &&
                             (String.IsNullOrEmpty(deptCode) || deptCode == c.出発地コード) &&
                             (String.IsNullOrEmpty(arrvCode) || arrvCode == c.到着地コード) &&
                             (String.IsNullOrEmpty(levelCode) || levelCode == d.クラスコード)
                         select new ScheduleInfo()
                         {
                             SheduleId = s.運行コード,
                             ScheduledDate = s.運行日,
                             CruiseId = s.便名,
                             DeptCode = c.出発地コード,
                             DeptName = p1.場所名,
                             DeptTime = c.出発時刻,
                             ArrvCode = c.到着地コード,
                             ArrvName = p2.場所名,
                             ArrvTime = c.到着時刻,
                             Price = r.基本料金,
                             Mile = r.運行マイル,
                             LevelCode = d.クラスコード,
                             LevelName = cl.クラス名,
                             Capacity = d.定員数,
                             Room = d.空き数,
                             AppliedPrice = (d.クラスコード == FirstClassCode) ?
                                             r.基本料金 + FirstExtraPrice : r.基本料金,
                             AppliedTax = ((d.クラスコード == FirstClassCode) ?
                                             r.基本料金 + FirstExtraPrice : r.基本料金) * TaxRatio
                         };
 */
