﻿using System;
using System.Linq;
using System.Transactions;
using System.Data;

using Adwk.Reservation.Dac;
using Adwk.Reservation.Data;


namespace Adwk.Reservation.Logic
{
    // 予約関連のロジックを含むビジネスコンポーネント
    // (予約の確認と削除)
    // 　このコンポーネントでは、予約を確認したのち、その情報をもとに
    // 予約の削除を行うので、予約データをステートとして一時的に保持する
    // ようにしました。よって、非静的クラスにしています。
    public class BLComp3 : IDisposable
    {
        // このコンポーネントが返すメッセージ
        const string msgInvalidReservationCode = "指定した予約コードが正しくありません(数字でなければなりません)。";
        const string msgNoReservation = "指定した予約は存在しません。";
        const string msgReservationExist = "指定した予約が見つかりました。";
        const string msgNotReferredReservation = "削除の前に、該当予約データを照会し確認してください。";
        const string msgCantDeleteReservation = "未乗船の予約ではないので、削除できません。";
        const string msgDeleteReservationSuccess = "予約を削除しました。";
        const string msgReservationDbAccessError = "システムエラーです。(0201)";
        const string smgScheduleDbAccessError = "システムエラーです。(0301)";
        const string msgOptimisticConCurrencyError = "システムエラーです。(0302)";
        const string msgDeleteReservationError = "システムエラーです。(0303)";
        const string msgDeleteReservationEtcError = "システムエラーです。(0305)";

        // このコンポーネントが維持するステート
        private AdwkReservationsEntities adwkRObjectContext;
        private 予約 currentReservation;  // 現在照会している予約

        // コンストラクター
        public BLComp3()
        {
            adwkRObjectContext = null;
            currentReservation = null;
        }

        // ObjectContextの準備
        private void PrepareObjectContext()
        {
            if (adwkRObjectContext == null)
            {
                adwkRObjectContext = new AdwkReservationsEntities();

                // ナビゲーションプロパティを参照時に自動的に読み込ませる
                adwkRObjectContext.ContextOptions.LazyLoadingEnabled = true; //EDMデザイナー既定値
            }
        }

        // 予約の照会(ReservationInfoとして返すバージョン)
        //   指定した予約コード(reservationCode)、氏名(lastnameとfirstName)が
        // 一致した場合のみ、予約データを返します。
        //   予約有無の確認なので、予約がなくともtrue(正常処理)を返します。
        //   予約がない場合は、引数は reservation は null値になります。
        public bool QueryReservation1ByRCode(
            string reservationCode, string lastName, string firstName,
            out ReservationInfo reservation,
            out AdwkRStatus stat,
            out string systemMessage)
        {
            reservation = null;
            currentReservation = null; //問い合わせごとに、照会中のキャッシュはクリア

            // 予約コードの検証(数値であるか)
            int rsvIntCode = 0;
            try
            {
                rsvIntCode = Convert.ToInt32(reservationCode);
            }
            catch
            {
                stat = AdwkRStatus.InvalidArgument;
                systemMessage = msgInvalidReservationCode;
                return false;
            }
            System.Diagnostics.Debug.Assert(rsvIntCode > 0);

            // ObjectContextの準備
            PrepareObjectContext();

            // 予約の照会
            // 搭乗済み r.乗船受付フラグ == "1" の予約データも返す
            var query = from r in adwkRObjectContext.予約
                        where r.予約コード == rsvIntCode &&
                              r.姓 == lastName && r.名 == firstName
                        select r;
            try
            {
                currentReservation = query.First();
            }
            catch (InvalidOperationException) //該当する予約無し
            {
                // 予約有無の確認なので、該当しない場合も、
                // 正常処理として true を返す。
                stat = AdwkRStatus.NoReservation;
                systemMessage = msgNoReservation;
                return true;
            }
            catch
            {
                stat = AdwkRStatus.UndefinedError;
                systemMessage = msgReservationDbAccessError;
                return false;
            }
            System.Diagnostics.Debug.Assert(currentReservation != null);

            //ReservationInfo型のデータを作成
            reservation = new ReservationInfo()
            {
                ReservationCode = currentReservation.予約コード.ToString(),
                MemberCode = currentReservation.会員コード,
                LastName = currentReservation.姓,
                FirstName = currentReservation.名,
                ScheduleCode = currentReservation.運行コード,
                ScheduleDate = currentReservation.運行予定明細.運行予定.運行日,
                LevelCode = currentReservation.クラスコード,
                LevelName = currentReservation.運行予定明細.クラス.クラス名,
                CruiseId = currentReservation.運行予定明細.運行予定.便名,
                DeptName = currentReservation.運行予定明細.運行予定.便情報.出発地の場所.場所名,
                DeptTime = currentReservation.運行予定明細.運行予定.便情報.出発時刻,
                ArrvName = currentReservation.運行予定明細.運行予定.便情報.到着地の場所.場所名,
                ArrvTime = currentReservation.運行予定明細.運行予定.便情報.到着時刻,
                AppliedPrice = currentReservation.適用料金,
                AppliedTax = currentReservation.適用料金消費税,
                BoardingFlag = currentReservation.乗船受付フラグ,
                BoardingDate = currentReservation.乗船日時,
                AppliedMile = currentReservation.実績マイル
            };

            // 正常終了
            stat = AdwkRStatus.Success;
            systemMessage = msgReservationExist;
            return true;
        }

        // 予約の照会([予約]エンティティとして返すバージョン)
        //   指定した予約コード(reservationCode)、氏名(lastnameとfirstName)が
        // 一致した場合のみ、予約データを返します。
        public bool QueryReservation2ByRCode(
            string reservationCode, string lastName, string firstName,
            out 予約 reservation,
            out AdwkRStatus stat,
            out string systemMessage)
        {
            reservation = null;
            currentReservation = null; //問い合わせごとに、照会中のキャッシュはクリア

            // 予約コードの検証(数値であるか)
            int rsvIntCode = 0;
            try 
            { 
                rsvIntCode = Convert.ToInt32(reservationCode); 
            }
            catch
            {
                stat = AdwkRStatus.InvalidArgument;
                systemMessage = msgInvalidReservationCode;
                return false;
            }
            System.Diagnostics.Debug.Assert(rsvIntCode > 0);

            // ObjectContextの準備
            PrepareObjectContext();

            // 予約の照会
            // 搭乗済み r.乗船受付フラグ == "1" の予約データも返す
            var query = from r in adwkRObjectContext.予約
                        where r.予約コード == rsvIntCode &&
                              r.姓 == lastName && r.名 == firstName
                        select r;
            try
            {
                currentReservation = query.First();
            }
            catch (InvalidOperationException) //該当する予約無し
            {
                // 予約有無の確認なので、該当しない場合も、
                // 正常処理として true を返す。
                stat = AdwkRStatus.NoReservation;
                systemMessage = msgNoReservation;
                return true;
            }
            catch
            {
                stat = AdwkRStatus.UndefinedError;
                systemMessage = msgReservationDbAccessError;
                return false;
            }
            System.Diagnostics.Debug.Assert(currentReservation != null);

            // 正常終了
            reservation = currentReservation;
            stat = AdwkRStatus.Success;
            systemMessage = msgReservationExist;
            return true;
        }

        // 予約の削除
        //   ここでは「照会したのちに削除する」という手順を呼び出し側が行うという
        // 前提なので、削除を指定した際には、そのデータが現在照会済みであるか
        // チェックしてから削除しています。
        public bool DeleteReservationByRCode(
            string reservationCode, 
            out AdwkRStatus stat, out string systemMessage)
        {
            // 予約コードの検証(数値であるか)
            int rsvIntCode = 0;
            try
            {
                rsvIntCode = Convert.ToInt32(reservationCode);
            }
            catch
            {
                stat = AdwkRStatus.InvalidArgument;
                systemMessage = msgInvalidReservationCode;
                return false;
            }
            System.Diagnostics.Debug.Assert(rsvIntCode > 0);

            // 予め照会確認しているか
            if (currentReservation == null ||
                currentReservation.予約コード != rsvIntCode)
            {
                stat = AdwkRStatus.NotReferredReservation;
                systemMessage = msgNotReferredReservation;
                return false;
            }
            System.Diagnostics.Debug.Assert( currentReservation != null );
            System.Diagnostics.Debug.Assert( adwkRObjectContext != null );

            // 削除可能であるか([乗船受付フラグ]が "0" であるか)を確認
            //   ここでは予約データの楽観的同時実行制御に基づいており、
            // キャッシュされた予約データのフラグを確認し、
            // 更新時に他から予約データのフラグが変更されてないか、
            // データ競合の有無を確認しています。
            if (currentReservation.乗船受付フラグ != AdwkCheckInFlag.NotCheckedIn)
            {
                stat = AdwkRStatus.CantDeleteReservation;
                systemMessage = msgCantDeleteReservation;
                return false;
            }

            // 削除処理の実行
            try
            {
                using (var ts = new TransactionScope(TransactionScopeOption.Required))
                {
                    // 空き数を戻す
                    運行予定明細 sinfo;
                    var query1 = from s in adwkRObjectContext.運行予定明細
                                 where s.運行コード == currentReservation.運行コード &&
                                       s.クラスコード == currentReservation.クラスコード
                                 select s;
                    try
                    {
                        // 有効な予約データに対する運用予定明細データなので見つかるはず
                        sinfo = query1.First();
                    }
                    catch //その他のアクセスエラー
                    {
                        stat = AdwkRStatus.UndefinedError;
                        systemMessage = smgScheduleDbAccessError;
                        return false;
                    }
                    sinfo.空き数++;

                    // 対象予約を削除としてマーク
                    // (簡単にするため、予約データ自体を削除)
                    adwkRObjectContext.予約.DeleteObject(currentReservation);

                    try
                    {
                        // 更新
                        adwkRObjectContext.SaveChanges();
                    }
                    catch (OptimisticConcurrencyException)
                    {
                        // 予約データの競合が発生した場合の例外
                        // 予約データは新規挿入したあと、他から[乗船受付フラグ]が変更される可能性があります。
                        // (同時に、[コメント]、[乗船日時]、[実績マイル]も更新されるが、
                        // これらは単独で更新されることはありません。
                        // (そのほか、予約されると他の列の内容は固定であることを前提にしています。)
                        // そのため、[乗船受付フラグ]においてデータ競合が発生する可能性があり
                        // その場合、OptimisticConcurrencyExceptionが発生します。
                        // EDMにおいて、[乗船受付フラグ]の同時実行モードは Fixed に設定されています。
                        // 他から削除されていた場合も、OptimisticConcurrencyExceptionが発生します。
                        stat = AdwkRStatus.OptimisticConcurrencyError;
                        systemMessage = msgOptimisticConCurrencyError;
                        return false;
                    }
                    catch
                    {
                        stat = AdwkRStatus.UndefinedError;
                        systemMessage = msgDeleteReservationError;
                        return false;
                    }

                    // 正常終了に投票(Vote)
                    ts.Complete();

                    // 正常終了の情報設定
                    stat = AdwkRStatus.Success;
                    systemMessage = msgDeleteReservationSuccess;
                }
            }
            catch
            {
                stat = AdwkRStatus.UndefinedError;
                systemMessage = msgDeleteReservationEtcError;
                return false;
            }

            // 正常終了
            System.Diagnostics.Debug.Assert(stat == AdwkRStatus.Success);
            System.Diagnostics.Debug.Assert(systemMessage != null);
            return true;
        }

        // このコンポーネントを使い終わった際の明示的な後処理
        public void Dispose()
        {
            currentReservation = null;
            if (adwkRObjectContext != null)
            {
                adwkRObjectContext.Dispose();
                adwkRObjectContext = null;
            }
        }
    }
}
